package com.sillycycle.bagleyd.ant2d;

/*
 * @(#)Ant2DBoxes.java
 *
 * Copyright 2010 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JPanel;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComboBox;

/**
 * The <code>Ant2DBoxes</code> class holds the Life configuration
 * for the <code>Ant2DDialog</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant2DBoxes extends JPanel {
	private static final long serialVersionUID = 42L;

	static int neighborIndexValue = Ant2DInterface.DEFAULT_NEIGHBOR_INDEX;
	static String ruleName = Ant2DRules.getRuleName(neighborIndexValue);
	transient Ant2DRules ant2dRules = new Ant2DRules(Ant2DRules.indexToNeighbor[neighborIndexValue]);

	ComboBox neighborComboBox = new ComboBox("Neighborhood Modes",
		Ant2DRules.neighborIndexStrings);

	public Ant2DBoxes() {
		JPanel panel = new JPanel();

		panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		neighborComboBox.setSelectedItem(neighborIndexValue);
		panel.add(neighborComboBox);
		add(BorderLayout.PAGE_START, panel);
		//pack();
		neighborComboBox.addActionListener(
			new NeighborComboBoxActionListener());
	}

	class NeighborComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int oldValue = neighborIndexValue;

			neighborIndexValue = neighborComboBox.getSelectedItem();
			if (neighborIndexValue != -1 && oldValue != neighborIndexValue) {
				ruleName = Ant2DRules.getRuleName(neighborIndexValue);
			}
		}
	}

	public void addActionListener(ActionListener l) {
		neighborComboBox.addActionListener(l);
		neighborIndexValue = neighborComboBox.getSelectedItem();

		ruleName = Ant2DRules.getRuleName(neighborIndexValue);
	}

	public void setNeighborIndex(int value) {
		Ant2DDialog.neighborIndexValue = value;
		neighborComboBox.setSelectedItem(value);
	}

	public int getNeighborIndex() {
		return neighborComboBox.getSelectedItem();
	}
}
