package com.sillycycle.bagleyd.ant2d;

/*
 * @(#)Ant2DCanvas.java
 *
 * Copyright 2010 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Random;
import javax.swing.JFrame;

import com.sillycycle.bagleyd.util.Cell2D;
import com.sillycycle.bagleyd.util.CellInfo;
import com.sillycycle.bagleyd.util.FiniteStateMachine;
import com.sillycycle.bagleyd.util.ModArith;
import com.sillycycle.bagleyd.util.Turmite2D;

/**
 * The <code>Ant2DCanvas</code> class does all the graphics for
 * <code>Ant2DFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant2DCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");

	static final byte MAX_STATES = 8;
	static final boolean MONO = false;
	static final String DATAFILE = "data.life";

	Random generator;
	int startSpeed = Ant2DInterface.DEFAULT_SPEED;
	int speed = Ant2DInterface.DEFAULT_SPEED;
	int columns = Ant2DInterface.DEFAULT_COLUMNS;
	int rows = Ant2DInterface.DEFAULT_ROWS;
	int ncols = columns;
	int nrows = rows;
	int minCol = columns / 2 - 1;
	int maxCol = columns / 2 + 1;
	int minRow = rows / 2 - 1;
	int maxRow = rows / 2 + 1;
	boolean randomOrient = false;
	boolean grid = true;
	boolean wrap = true;
	boolean visibleCell = true;
	boolean useTable = false;
	boolean eyes = true;
	boolean sharpTurn = false;
	boolean truchet = true;
	boolean focus = true;
	boolean mouseDown = false;
	int xs = Ant2DInterface.DEFAULT_CELLSIZE;
	int ys = Ant2DInterface.DEFAULT_CELLSIZE;
	int neighbors =
		Ant2DRules.indexToNeighbor[Ant2DInterface.DEFAULT_NEIGHBOR_INDEX];
	Color initialCellColor = Color.red;
	Color gridColor = Color.gray;
	Color background = Color.black;
	Cell2D cellPl = null;
	transient Capturer capturer;
	JFrame frame = null;
	boolean debug = false;
	String name = "newbie";
	Ant2DMatrix field = null;
	int gen = 0;
	int nCells = 0, cCells = 0;
	String stringSave = null;
	transient Ant2DRules param = null;
	transient Turmite2D turmite = null;
	int rule = Ant2DRules.MIN_RULE;
	static int[] newDir5 = {0, 2, 1, 4, 3};

	private synchronized void pulse(boolean draw) {
		int tapePos, newPosition, statePos, col, row;
		byte color, newColor;
		short oldDir, arrow, newDir;
		int otherDir = 0;
		FiniteStateMachine status;

		cCells = 0;
		col = turmite.getColumn();
		row = turmite.getRow();
		tapePos = getPosition(col, row, ncols);
		color = field.getCell(tapePos); /* read tape */
		statePos = color + turmite.getState() * param.getNColors();
		status = param.machine[statePos];
		newColor = status.getColor();
		if (debug)
			System.out.println("statePos " + statePos +
				", newColor " + newColor);
		CellInfo info = new CellInfo(tapePos, newColor, (byte) 0);

		drawCell(info);
		field.setCell(tapePos, newColor); /* write on tape, ncells updated */
		checkAliveBounds(col, row);
		/* Find direction of Bees or Ants. */
		/* Translate relative direction to actual direction */
		oldDir = turmite.getDirection();
		arrow =	status.getDirection();
		newDir = heading(oldDir, arrow);
/*System.out.println("oldDir " + oldDir / 72 + ", arrow" + arrow + ", newDir" + newDir / 72);
int orient = ((row & 1) * 2 + col) % 4;
System.out.println("orient " + orient);*/
		turmite.setDirection(newDir);

		if (!useTable) {
			Graphics g = this.getGraphics();
			int truchetState = 0;

			switch (neighbors) {
			case 3:
			    int chgDir = ((720 - arrow) % 360);
			    truchetState = ((chgDir == 240) ?
				    (2 + newDir / 120) % 3 :
				    (1 + newDir / 120) % 3);
			    break;
			case 4:
			    {
				int a = oldDir / 180;
				int b = newDir / 180;
				truchetState = (((a != 0 && b == 0) || (b != 0 && a == 0)) ? 1 : 0);
				break;
			    }
			case 5:
			    {
				int a, b;
				a = (oldDir / 72) % 5;
				b = (newDir / 72) % 5;
				// Fixed by trial and error...
				if (sharpTurn) {
				    truchetState = (a + b + 4) % 5;
				    if ((a == 2 && b == 0) || (a == 3 && b == 0)) {
					truchetState = (truchetState + 1) % 5;
				    }
				    if ((a == 0 && b == 1) || (a == 1 && b == 3) ||
					    (a == 4 && b == 4)) {
					truchetState = (truchetState + 2) % 5;
				    }
				    if (a == 2 && b == 2) {
					truchetState = (truchetState + 3) % 5;
				    }
				    if (a == 3 && b == 3) {
					truchetState = (truchetState + 4) % 5;
				    }
				} else {
				    truchetState = (a + b) % 5;
				    if (a == 2 && b == 4) {
					truchetState = (truchetState + 1) % 5;
				    }
				    if (a == 3 && b == 1) {
					truchetState = (truchetState + 3) % 5;
				    }
				    if ((a == 0 && b == 2) || (a == 1 && b == 4) ||
					    (a == 3 && b == 2) || (a == 4 && b == 0)) {
					truchetState = (truchetState + 4) % 5;
				    }
				}
				otherDir = newDir;
				if ((a == 0 && b == 2) || (a == 0 && b == 4) ||
					(a == 1 && b == 4) || (a == 2 && b == 2) ||
					(a == 2 && b == 4) || (a == 3 && b == 2) ||
					(a == 4 && b == 2) || (a == 4 && b == 4)) {
				    otherDir = ((newDir + 216) % 360);
				}
				if ((a == 1 && b == 0) || (a == 2 && b == 0)
					|| (a == 3 && b == 0) || (a == 4 && b == 0)) {
				    otherDir = ((newDir + 288) % 360);
				}
				break;
			    }
			case 6:
			    if (sharpTurn) {
				boolean a, b;
				a = (((360 + newDir - oldDir) % 360) == 240);
				/* should be some way of getting rid of the init_dir dependency... */
				b = (arrow % 120 == 0);
				truchetState = (((a && !b) || (b && !a)) ? 1 : 0);
			    } else {
				int a, b;
				a = (oldDir / 60) % 3;
				b = (newDir / 60) % 3;
				truchetState = (a + b + 1) % 3;
			    }
			    break;
			default:
			    truchet = false;
			    break;
			}
			field.setTruchet(tapePos, truchetState);
			if (draw && truchet)
				cellPl.drawTruchet(g,
					(newColor == 0) ? Color.white : Color.black,
					truchetState, sharpTurn,
					col, row);
		}
		turmite.setState(status.getNext());
		/* Allow step first then turn */
		oldDir = ((arrow < totalAngle(neighbors)) ?
			newDir : oldDir);
		if (debug)
			System.out.println("newDir " + newDir +
				", oldDir " + oldDir +
				", col " + col +
				", row " + row);
		newPosition = positionOfNeighbor(oldDir, col, row);
		int newCol = getColumn(newPosition);
		int newRow = getRow(newPosition);
		turmite.setPosition(newCol, newRow);
		if (debug)
			System.out.println("neighbors " +
				neighbors + ", newDir " + newDir +
				", newCol " + newCol +
				", newRow " + newRow);
		if (neighbors == 5) {
			newDir = (short) otherDir;
		}
		drawTurmite(newDir, newCol, newRow);
		gen++;
	}

	void turmitePulse(boolean draw, boolean fast) {
		pulse(draw);
		if (draw && (!fast || (gen & 63) == 0)) {
			((Ant2DFrame) frame).updateGen(gen);
			((Ant2DFrame) frame).updateCells(nCells);
		}
	}

	static int totalAngle(int neighbors) {
		if (neighbors == 7)
			return 357; // 360 not divisible by 7
		return 360;
	}

	public Ant2DCanvas() {
		generator = new Random(System.nanoTime());
		capturer = new Capturer();
		inits();
	}

	final void inits() {
		initTable();
		resizeField();
		clearField(true);
		prepopulate(false);
	}

	void initTable() {
		param = new Ant2DRules(neighbors/*,
			Ant2DRules.getRuleName(getNeighborIndex())*/);
		param.initMachine();
		param.parseRule(neighbors, rule, useTable, sharpTurn);
	}

	void ruleChange() {
		if ((Ant2DFrame) frame != null) {
			param = new Ant2DRules(neighbors/*,
				Ant2DRules.getRuleName(getNeighborIndex())*/);
		}
	}

	public int getNeighborIndex() {
		return Ant2DRules.neighborToIndex(neighbors);
	}

	static boolean polygonChange(int oldNeighborIndex, int newNeighborIndex) {
		if (oldNeighborIndex != newNeighborIndex) {
			int oldPolygon = Ant2DRules.neighborIndexToPolygon[oldNeighborIndex];
			int newPolygon = Ant2DRules.neighborIndexToPolygon[newNeighborIndex];

			return (oldPolygon != newPolygon);
		}
		return false;
	}

	int getPolygon() {
		return Ant2DRules.neighborToPolygon(neighbors);
	}

	void setNeighbors(int neighborIndex) {
		neighbors = Ant2DRules.indexToNeighbor[neighborIndex];
	}

	int getRealColumns(int cols) {
		int polygon = getPolygon();
		int realColumns = cols;

		if (polygon == 3 || polygon == 6) {
			if ((realColumns & 1) == 1)
				realColumns--;
		} else if (polygon == 5) {
			realColumns -= realColumns % 4;
		}
		return realColumns;
	}

	int getRowsFromColumns(int cols) {
		int polygon = getPolygon();
		int calculatedRows = 16 * cols / 25;

		if ((polygon == 3 || polygon == 5 || polygon == 6) &&
				((calculatedRows & 1) == 1)) {
			calculatedRows--;
		}
		return calculatedRows;
	}

	public int getColumns() {
		return this.ncols;
	}

	public int getRows() {
		return this.nrows;
	}

	public void speedUp() {
		speed += 10;
		if (speed > Ant2DInterface.MAX_SPEED)
			speed = Ant2DInterface.MAX_SPEED;
		((Ant2DFrame) frame).setDelay(Ant2DInterface.MAX_SPEED +
			Ant2DInterface.MIN_SPEED - 1 - speed);
	}

	public void slowDown() {
		speed -= 10;
		if (speed < Ant2DInterface.MIN_SPEED)
			speed = Ant2DInterface.MIN_SPEED;
		((Ant2DFrame) frame).setDelay(Ant2DInterface.MAX_SPEED +
			Ant2DInterface.MIN_SPEED - 1 - speed);
	}

	void resizeField() {
		int newWidth = getSize().width;
		int newHeight = getSize().height;
		int polygon = getPolygon();
		int newI = 0, newJ = 0;

		capturer.clearCapture();
		int tempCols = Ant2DFrame.getColumns();
		ncols = getRealColumns(tempCols);
		nrows = getRowsFromColumns(tempCols);
		if (newWidth == 0)
			newWidth = Ant2DInterface.DEFAULT_CELLSIZE *
				ncols + 4;
		if (newHeight == 0)
			newHeight = Ant2DInterface.DEFAULT_CELLSIZE *
				nrows + 4;
		cellPl = new Cell2D(polygon, ncols, nrows,
			newWidth, newHeight, false);
		if (debug) {
			System.out.println("Polygon " + polygon);
			System.out.println("Resize " + neighbors);
			System.out.println("ncols " + ncols + ", " + nrows);
			System.out.println("cols " + columns + ", " + rows);
		}
		if (field == null) {
			columns = ncols;
			rows = nrows;
			field = new Ant2DMatrix(ncols, nrows);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;
			if ((Ant2DFrame) frame != null) {
				((Ant2DFrame) frame).initPopup();
			}
		} else if (columns != ncols || rows != nrows) {
			if (polygon == 3) {
				if (((minCol + minRow) & 1) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol--;
					}
				}
			} else if (polygon == 5) {
				int mod = (minCol + 2 * minRow) % 4;

				if ((mod / 2) == 1) {
					if ((minRow & 1) == 1) {
						minRow--;
					} else {
						minCol -= 2 ;
					}
				}
				minCol -= (mod & 1);
			} else if ((polygon == 6) && ((minRow & 1) == 1)) {
				minRow--;
			}

			int fieldX = maxCol - minCol + 1;
			int fieldY = maxRow - minRow + 1;
			Ant2DMatrix cage = new Ant2DMatrix(fieldX, fieldY);

			for (int j = minRow; j <= maxRow; j++)
			for (int i = minCol; i <= maxCol; i++) {
				int oldP = getPosition(i, j, columns);
				byte value = field.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i - minCol, j - minRow,
						fieldX);
					cage.setCell(p, value);
					if (!useTable && field.trucheted(p)) {
						cage.setTruchet(p, field.getTruchet(oldP));
					}
				}
			}
			if (turmite != null) {
				newI = turmite.getColumn() - minCol;
				newJ = turmite.getRow() - minRow;
			}

			columns = ncols;
			rows = nrows;
			field = new Ant2DMatrix(ncols, nrows);
			gen = 0;
			minCol = ncols / 2 - 1;
			maxCol = ncols / 2 + 1;
			minRow = nrows / 2 - 1;
			maxRow = nrows / 2 + 1;
			nCells = 0;

			int offsetX = minCol + 1 - fieldX / 2;
			int offsetY = minRow + 1 - fieldY / 2;

			if (polygon == 3) {
				if (((offsetX + offsetY) & 1) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX--;
					}
				}
			} else if (polygon == 5) {
				int mod = (offsetX + 2 * offsetY) % 4;

				if ((mod / 2) == 1) {
					if ((offsetY & 1) == 1) {
						offsetY--;
					} else {
						offsetX -= 2 ;
					}
				}
				offsetX -= (mod & 1);
			} else if ((polygon == 6) && ((offsetY & 1) == 1)) {
				offsetY--;
			}
			for (int j = 0; j < fieldY; j++)
			if ((j + offsetY >= 0) && (j + offsetY <= nrows - 1))
			for (int i = 0; i < fieldX; i++)
			if ((i + offsetX >= 0) && (i + offsetX <= ncols - 1)) {
				int oldP = getPosition(i, j, fieldX);
				byte value = cage.getCell(oldP);

				if (value != 0) {
					int p = getPosition(i + offsetX, j + offsetY,
						ncols);
					field.setCell(p, value);
					if (!useTable && cage.trucheted(p)) {
						field.setTruchet(p, cage.getTruchet(oldP));
					}
					nCells++;
				}
			}
			if (turmite != null) {
				turmite.setPosition(newI + offsetX, newJ + offsetY);
			}
			if ((Ant2DFrame) frame != null) {
				((Ant2DFrame) frame).updateGen(gen);
				((Ant2DFrame) frame).updateCells(nCells);
				((Ant2DFrame) frame).initPopup();
			}
		}
		if (grid) {
			eraseFrame();
			drawGrid();
		}
	}

	void xorCell(Graphics g, CellInfo info, int k) {
		int col = getColumn(info);
		int row = getRow(info);
		byte state = info.getCell();
		Color color;

		if (state == 0) {
			color = background;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorWheel(state);
			}
		}
		if ((k & 1) == 0) // Simulate Xor
			color = new Color(255 - color.getRed(),
				255 - color.getGreen(),
				255 - color.getBlue());
		cellPl.drawCell(g, false, color,
			(grid) ? gridColor : background, col, row);
	}

	// Should make some attempt at minimizing size if on a border...
	void printField(StringBuilder sb) {
		String space = " ";
		String newline = NL;
		String pound = "#";
		int polygon = getPolygon();
		boolean firstPoint = true;
		int i, j, p;
		int startX = 0, startY = 0;
		int endX = 0, endY = 0;
		byte state;

		sb.append(pound).append("Dim").append(space)
			.append(2).append(newline);
		sb.append(pound).append("Neighbors").append(space)
			.append(neighbors).append(newline);
		sb.append(pound).append("Rule").append(space)
			.append(rule).append(newline);
		sb.append(pound).append("Cells").append(space)
			.append(nCells).append(newline);
		sb.append(pound).append("Name").append(space)
			.append(name).append(newline);
		/*sb.append(pound).append("Sig").append(space)
			.append("An A0 <> D1 Dn").append(newline);*/
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++) {
				p = getPosition(i, j);
				if (field.getCell(p) != 0) {
					if (firstPoint) {
						startX = i;
						startY = j;
						endX = i;
						endY = j;
						firstPoint = false;
					} else {
						if (i < startX)
							startX = i;
						if (j < startY)
							startY = j;
						if (i > endX)
							endX = i;
						if (j > endY)
							endY = j;
					}
				}
		}
		sb.append(pound).append("Width").append(space)
			.append(endX - startX).append(1).append(newline);
		sb.append(pound).append("Height").append(space)
			.append(endY - startY).append(1).append(newline);
		sb.append(pound).append("P").append(space)
			.append(startX - ncols / 2).append(space)
			.append(startY - nrows / 2).append(newline);
		/*sb.append(pound).append("P").append(space)
			.append(0).append(space)
			.append(0).append(newline);*/
		switch (polygon) {
		case 3:
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
			break;
		case 5:
			startX -= (startX + 2 * startY) % 4;
			break;
		case 6:
			int append = (endY - startY + 1) / 2;
			int low = endX + append - startX;
			int high = 0;
			int offset = 0;

			for (j = startY; j <= endY; j++) {
				for (i = startX + offset; i <= endX + append; i++) {
					p = getPosition(i, j);
					state = field.getCell(p);
					if (state != 0) {
						if (low > i - startX - offset)
							low = i - startX - offset;
						if (high < i - startX - offset )
							high = i - startX - offset;
					}
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
			}
			append = high;
			offset = low;
			for (j = startY; j <= endY; j++) {
				for (int k = j; k < endY; k++) {
					sb.append(space);
				}
				for (i = startX + offset; i <= startX + append; i++) {
					if (i > startX + offset)
						sb.append(space);
					p = getPosition(i, j);
					sb.append(field.getCell(p));
				}
				if ((j & 1) == 1) {
					offset--;
					append--;
				}
				sb.append(newline);
			}
			return;
		default:
		}
		for (j = startY; j <= endY; j++) {
		for (i = startX; i <= endX; i++) {
				p = getPosition(i, j);
				sb.append(field.getCell(p));
		}
		sb.append(newline);
		}
	}

	void clearField(boolean draw) {
		int i, j, p;

		for (j = minRow; j <= maxRow; j++)
		for (i = minCol; i <= maxCol; i++) {
				p = getPosition(i, j);
				field.clearCell(p);
		}
		/*// This will flash
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++)
				cellPl.drawCell(g, false, background, null,
					i, j);*/
		if (draw)
			eraseFrame();
		gen = 0;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		nCells = 0;
		if (draw && (Ant2DFrame) frame != null) {
			Graphics g = this.getGraphics();

			((Ant2DFrame) frame).updateGen(gen);
			((Ant2DFrame) frame).updateCells(nCells);
			drawFrame(g, true);
			if (grid) {
				drawGrid();
			}
		}
	}

	void prepopulate(boolean draw) {
		int col = ncols / 2;
		int row = nrows / 2;
		int polygon = getPolygon();
		int orient = 0;

		if (polygon == 3) {
			if (((col + row) & 1) == 0)
				col--;
		} else if (polygon == 5) {
			int mod = (col + 2 * minRow) % 4;
			col -= mod + 1;
		}
		if (randomOrient) {
			if (polygon == 5) {
				col += generator.nextInt(4) - 1;
			} else {
				orient = generator.nextInt(polygon);
			}
		}
		orient = (short) (orient * totalAngle(neighbors) / neighbors);
		turmite = new Turmite2D(col, row, (short) orient, (byte) 0);

		if (draw)
			drawTurmite((short) orient, col, row);
	}

	void showMessage(String msg) {
		((Ant2DFrame) frame).showMessage(msg);
	}

	void setCell(int col, int row, byte state) {
		if (col >= 0 && col < ncols &&
				row >= 0 && row < nrows) {
			setCell(col, row, state, visibleCell);
		}
	}

	/* rules are not reflective symmetric so must know */
	public boolean mirror(int orient) {
		int polygon = getPolygon();

		if (polygon == 4) {
			if (orient >= 4) {
				if ((orient % 4 >= 2) &&
						((orient & 1) == 1))
					return true;
				if ((orient % 4 < 2) &&
						((orient & 1) == 0))
					return true;
			} else {
				if ((orient % 4 >= 2) &&
						((orient & 1) == 0))
					return true;
				if ((orient % 4 < 2) &&
						((orient & 1) == 1))
					return true;
			}
		} else if ((polygon == 6) && (orient >= 6)) {
			return true;
		}
		return false;
	}

	private void setOrientCell(int x, int y,
			int startX, int startY, byte state,
			int orient, boolean draw) {
		int i = x, j = y, temp;
		int polygon = getPolygon();

		switch (polygon) {
		case 3:
			if (orient >= 6) {
				i = -i;
			}
			switch (orient % 6) {
			default:
			case 0:
				break;
			case 1:
				temp = j;
				j = ModArith.ceil2(temp + i);
				i = ModArith.ceil2(i - temp + 1) - temp;
				break;
			case 2:
				temp = j;
				j = ModArith.ceil2(temp - i);
				i = temp - ModArith.ceil2(1 - i - temp);
				break;
			case 3:
				i = -i;
				j = -j + 1;
				break;
			case 4:
				temp = j;
				j = 1 - ModArith.ceil2(temp + i);
				i = temp - ModArith.ceil2(i - temp + 1);
				break;
			case 5:
				temp = j;
				j = 1 - ModArith.ceil2(temp - i);
				i = ModArith.ceil2(1 - i - temp) - temp;
				break;
			}
			break;
		case 4:
			if (orient >= 4) {
				temp = i;
				i = j;
				j = temp;
			}
			if (orient % 4 >= 2) {
				i = -i;
			}
			if ((orient & 1) == 1) {
				j = -j;
			}
			break;
		case 5:
			int mod = ModArith.mod(i + 2 * j, 4);
			int ti = j * 2;
			int tj = ModArith.div(i, 2);
			if (mod == 1 || mod == 2)
				ti++;
			switch (orient % 4) {
			case 1:
				i = -ti - 1;
				j = tj;
				break;
			case 2:
				i = -i - 1;
				j = -j - 1;
				break;
			case 3:
				i = ti;
				j = -tj - 1;
				break;
			default:
				break;
			}
			if (orient >= 4) {
				i = -i + 1;
				mod = ModArith.mod(i + 2 * j, 4);
				if (mod == 0)
					i++;
				if (mod == 1)
					i--;
			}
			break;
		case 6:
			if (orient >= 6) {
				temp = i;
				i = j;
				j = temp;
			}
			switch (orient % 6) {
			default:
			case 0:
				break;
			case 1:
				temp = j;
				j = temp - i;
				i = temp;
				break;
			case 2:
				temp = -i;
				i = temp + j;
				j = temp;
				break;
			case 3:
				i = -i;
				j = -j;
				break;
			case 4:
				temp = -j;
				j = temp + i;
				i = temp;
				break;
			case 5:
				temp = i;
				i = temp - j;
				j = temp;
				break;
			}
			break;
		default:
			break;
		}
		if (polygon == 5) {
			temp = startX - (startX + startY * 2) % 4;
			i += temp;
		} else
			i += startX;
		if (polygon == 6) {
			if ((startX & 1) == 1 && j >= 0 || j > 0) {
				i += (((startX & 1) == 1) ? -j / 2 :
					-(j - 1) / 2);
			} else {
				i += 1 + (((startX & 1) == 1) ? -(j + 1) / 2 :
					-j / 2);
			}
		}
		j += startY;
		if (polygon == 3 &&
				(((startX + startY + 1) & 1) == 1)) {
			i++;
		}
		if (i >= 0 && i < ncols && j >= 0 && j < nrows)
			setCell(i, j, state, draw);
	}

	public void displayForm(byte[][] data, boolean draw) {
		byte state;
		int orient = 0;
		int polygon = getPolygon();
		int startX = ncols / 2 - data[0].length / 2;
		int startY = nrows / 2 - data.length / 2;

		if (polygon == 5) {
			startX -= (startX + startY * 2) % 4;
		} else if (polygon == 6 && ((startX + startY) & 1) == 0) {
			startY -= 1;
		}
		if (randomOrient) {
			if (polygon == 5) {
				startX += generator.nextInt(4) - 1;
			} else {
				orient = generator.nextInt(polygon);
			}
		}
		for (int j = 0; j < data.length; j++) {
			for (int i = 0; i < data[0].length; i++) {
				state = data[i][j];
				if (state != 0) {
					setOrientCell(i, j, startX, startY, state, orient, draw);
				}
			}
		}
	}

	public void getAnt() {
		this.stringSave = ((Ant2DFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getAntString();
	}

	public void getAntFile() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder fileData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				fileData.append(lineData).append(NL);
			this.stringSave = fileData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file to read");
				return;
			}
			getAntString();
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getAntString() {
		try {
			String space = " ";
			String newline = NL;
			String pound = "#";
			int dim = 0, neighborhood, cells;
			int width = 5, height = 5;
			int offsetX = 0, offsetY = 0;
			int i = 0;
			String ruleString = "";
			String[] lineToken = this.stringSave.split(newline);
			if (lineToken.length == 0)
				return;
			String lineData = lineToken[i];
			String[] wordToken = null;

			try {
				while (i < lineToken.length) {
					wordToken = lineData.split(space);
					if (wordToken.length > 1 &&
							(pound + "Dim").equals(wordToken[0]))
						break;
					if (++i == lineToken.length)
						return;
					lineData = lineToken[i];
				}
				if (wordToken != null && wordToken.length > 1)
					dim = Integer.parseInt(wordToken[1]);
				if (dim != 2) {
					showMessage(DATAFILE +
						" corrupted: dim " + dim);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: dim not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Neighbors
				neighborhood = Integer.parseInt(wordToken[1]);
				if (Math.abs(neighborhood) <= 2) {
					showMessage(DATAFILE +
						" corrupted: neighbors " + neighborhood);
					neighborhood = this.neighbors;
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: neighbors not an integer");
				neighborhood = this.neighbors;
			}
			++i;
			// #Rule
			//ruleString = Ant2DRules.parseRule(neighbors, wordToken[1]);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Cells
				cells = Integer.parseInt(wordToken[1]);
				if (cells < 0) {
					showMessage(DATAFILE +
						" corrupted: cells " + cells);
				}
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: cells not an integer");
			}
			lineData = lineToken[++i];
			// #Name
			name = lineData.substring(lineData.indexOf(space) + 1);
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Width
				width = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: width not an integer");
			}
			lineData = lineToken[++i];
			try {
				wordToken = lineData.split(space);
				// #Height
				height = Integer.parseInt(wordToken[1]);
			} catch (NumberFormatException e) {
				showMessage(DATAFILE +
					" corrupted: height not an integer");
			}
			lineData = lineToken[++i];
			wordToken = lineData.split(space);
			// #P
			if (wordToken.length > 1)
				offsetX = Integer.parseInt(wordToken[1]);
			if (wordToken.length > 2)
				offsetY = Integer.parseInt(wordToken[2]);
			//((Ant2DFrame) frame).callback(
			//	Ant2DInterface.ACTION_RESTORE);
			cells = scanField(lineToken, i, offsetX, offsetY);
			drawField(true);
			System.out.println(DATAFILE +
				": neighbors " + neighborhood +
				", rule " + ruleString +
				", cells " + cells +
				", name " + name +
				", width " + width +
				", height " + height +
				", offsetX " + offsetX +
				", offsetY " + offsetY);
		} catch (NumberFormatException e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading");
		}
	}

	public void writeAnt() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		((Ant2DFrame) frame).writeTextArea(sb.toString());
	}

	public void writeAntFile() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.println(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	public void printAnt() {
		StringBuilder sb = new StringBuilder();

		printField(sb);
		System.out.println(sb);
	}

	void eraseFrame(Graphics g) {
		g.setColor(background);
		g.fillRect(0, 0, getSize().width, getSize().height);
	}

	void eraseFrame() {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			eraseFrame(g);
		} finally {
			g.dispose();
		}
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		Rectangle r = cellPl.boundRect(ncols, nrows);

		if (g == null || r == null)
			return;
		g.setColor(background.darker().darker());
		g.fillRect(0, 0, getSize().width, r.y);
		g.fillRect(0, r.y, r.x, r.height - r.y);
		g.fillRect(0, r.height, getSize().width, r.y);
		g.fillRect(r.width, r.y,
			getSize().width - r.width, r.height - r.y);
		draw3DFrame(show && focus, r.x, r.y,
			r.width - r.x, r.height - r.y);
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	boolean checkAliveBounds(int i, int j) {
		boolean edge = false;

		if (minCol == 0 || maxCol == ncols - 1 ||
				i == 0 || i == ncols - 1) {
			minCol = 0;
			maxCol = ncols - 1;
			edge = true;
		} else if (i <= minCol) {
			minCol = i - 1;
		} else if (i >= maxCol) {
			maxCol = i + 1;
		}
		if (minRow == 0 || maxRow == nrows - 1 ||
				j == 0 || j == nrows - 1) {
			minRow = 0;
			maxRow = nrows - 1;
			edge = true;
		} else if (j <= minRow) {
			minRow = j - 1;
		} else if (j >= maxRow) {
			maxRow = j + 1;
		}
		return edge;
	}

	int scanField(String[] lineToken, int stringIndex,
			int offsetX, int offsetY) {
		int polygon = getPolygon();
		String lineData;
		int m = stringIndex, count = 0;
		int i, j, n, p, startX, startY, row;
		String space = " ";

		startX = offsetX + ncols / 2;
		startY = offsetY + nrows / 2;
		if (polygon == 3) {
			if (((startX + startY) & 1) == 1) {
				startX--;
			}
		} else if (polygon == 5) {
			startX -= (startX + 2 * startY) % 4;
		}
		try {
			row = 0;
			for (j = startY; ++m < lineToken.length; j++) {
				lineData = lineToken[m];
				if (lineData.length() == 0) {
					break; // allows flexibility
				}
				n = 0;
				i = startX;
				if (polygon == 3) {
					if (((startX + startY) & 1) == 1)
						i--;
				} else if (polygon == 6) {
					i -= (row + (startY & 1)) / 2;
				}
				while (n < lineData.length()) {
					byte oldCell, newCell;
					if (space.charAt(0) == lineData.charAt(n) ||
							i < 0 || i >= ncols ||
							j < 0 || j >= nrows) {
						n++;
						continue;
					}
					p = getPosition(i, j);
					newCell = (byte) (lineData.charAt(n) - '0');
					if (newCell != 0) {
						count++;
					}
					oldCell = field.getCell(p);
					if (oldCell != newCell) {
						field.setCell(p, newCell);
						CellInfo info = new CellInfo(
							p, newCell, (byte) 0);
						drawCell(info);
						if (newCell != 0 && oldCell == 0) {
							nCells++;
							checkAliveBounds(i, j);
						} else if (newCell == 0 && oldCell != 0) {
							nCells--;
						}
						((Ant2DFrame) frame).updateCells(nCells);
					}
					i++;
					n++;
				}
				row++;
			}
		} catch (Exception e) {
			e.printStackTrace();
			showMessage("Corrupt input found when reading (scanField)");
		}
		return count;
	}

	void drawField(boolean draw) {
		int i, j, p;
		byte state;

		nCells = 0;
		minCol = ncols / 2 - 1;
		maxCol = ncols / 2 + 1;
		minRow = nrows / 2 - 1;
		maxRow = nrows / 2 + 1;
		for (j = 0; j < nrows; j++)
		for (i = 0; i < ncols; i++) {
				p = getPosition(i, j);
				state = field.getCell(p);
				if (state != 0) {
					CellInfo info = new CellInfo(
						p, state, (byte) 0);
					if (draw)
						drawCell(info);
					nCells++;
					checkAliveBounds(i, j);
				}
				if (!useTable && truchet && field.trucheted(p)) {
					Graphics g = this.getGraphics();
					int truchetState = field.getTruchet(p);

					cellPl.drawTruchet(g,
						(state == 0) ? Color.white : Color.black,
						truchetState, sharpTurn,
						i, j);
				}
			}
		if (draw && (Ant2DFrame) frame != null) {
			if (grid) {
				drawGrid();
			}
			((Ant2DFrame) frame).updateCells(nCells);
			capturer.show();
		}
		if (draw) {
			drawTurmite(turmite.getDirection(),
				turmite.getColumn(), turmite.getRow());
		}
	}

	void redrawField() {
		eraseFrame();
		drawFrame(true);
		drawField(true);
	}

	public Rectangle getSelectedRectangle() {
		return capturer.getRectangle();
	}

	void banner(String msg) {
		((Ant2DFrame) frame).displayMessage(msg);
	}

	void drawGrid() {
		Graphics g = this.getGraphics();
		Color color;

		if (g == null)
			return;
		if (grid)
			color = Color.gray;
		else
			color = background;
		g.setColor(color);
		for (int j = 0; j < nrows; j++)
		for (int i = 0; i < ncols; i++) {
				cellPl.drawCell(g, false, null, null, i, j);
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		drawFrame(focus);
		drawField(true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}

	int getColumn(int position) {
		return position % this.ncols;
	}

	int getRow(int position) {
		return position / this.ncols;
	}

	static int getColumn(int position, int ncols) {
		return position % ncols;
	}

	static int getRow(int position, int ncols) {
		return position / ncols;
	}

	int getPosition(int col, int row) {
		return getPosition(col, row, this.ncols);
	}

	static int getPosition(int col, int row, int ncols) {
		return row * ncols + col;
	}

	int getColumn(CellInfo info) {
		return getColumn(info.getPosition());
	}

	int getRow(CellInfo info) {
		return getRow(info.getPosition());
	}

	void setPosition(CellInfo info, int col, int row) {
		info.setPosition(getPosition(col, row));
	}

	short heading(short direction, short turn) {
		short oldDirection = direction;
		short newDirection;
		short chgDir;

		if (neighbors == 5) {
			chgDir = (short) (newDir5[(turn / 72) % neighbors] * 72);
		} else {
			chgDir = (short) ((2 * totalAngle(neighbors) - turn) %
				totalAngle(neighbors));
		}
		if (neighbors == 5) {
			oldDirection = (short) (newDir5[oldDirection / 72] * 72);
		}
		newDirection = (short) ((chgDir + oldDirection) %
			totalAngle(neighbors));
		if (debug)
			System.out.println("oldDirection " + oldDirection +
				", turn " + turn +
				", chgDir " + chgDir +
				", newDirection " + newDirection);
		return newDirection;
	}

	int[] getNeighborArray(int col, int row) {
		int p;
		int i, n = Math.abs(neighbors);
		int[] c = new int[n];

		if ((col == 0 || row == 0 ||
				col == ncols - 1 || row == nrows - 1) &&
				!wrap) {
			return c;
		}
		for (i = 0; i < n; i++) {
			p = positionOfNeighbor(i, col, row);
			c[i] = field.getCell(p); // non-totalistic
		}
		return c;
	}

	int positionOfNeighbor(int cell, int i, int j) {
		int polygon = getPolygon();
		int col = i, row = j;
		int dir = cell; // * (360 / Math.abs(neighbors));

		if (neighbors < 0) { /* -6 or -9 */
			dir = (dir >= 180) ? dir - 180 : dir + 180;
		}
		switch (polygon) {
		case 3:
			/* TRI */
			if (((col + row) & 1) == 0) {	/* up */
				switch (dir) {
				case 0: /* close */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 20:
				case 30:
				case 40:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 60: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					switch (col) {
					case 0:
						col = ncols - 2;
						break;
					case 1:
						col = ncols - 1;
						break;
					default:
						col = col - 2;
						break;
					}
					break;
				case 80:
				case 90:
				case 100:
					switch (col) {
					case 0:
						col = ncols - 2;
						break;
					case 1:
						col = ncols - 1;
						break;
					default:
						col = col - 2;
						break;
					}
					break;
				case 120: /* close */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 140:
				case 150:
				case 160:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 180: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 200:
				case 210:
				case 220:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 240: /* close */
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 260:
				case 270:
				case 280:
					if (col + 1 == ncols)
						col = 1;
					else if (col + 2 == ncols)
						col = 0;
					else
						col = col + 2;
					break;
				case 300: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					if (col + 1 == ncols)
						col = 1;
					else if (col + 2 == ncols)
						col = 0;
					else
						col = col + 2;
					break;
				case 320:
				case 330:
				case 340:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
			} else {	/* down */
				switch (dir) {
				case 0: /* close */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 20:
				case 30:
				case 40:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 60: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					if (col + 1 == ncols)
						col = 1;
					else if (col + 2 == ncols)
						col = 0;
					else
						col = col + 2;
					break;
				case 80:
				case 90:
				case 100:
					if (col + 1 == ncols)
						col = 1;
					else if (col + 2 == ncols)
						col = 0;
					else
						col = col + 2;
					break;
				case 120: /* close */
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 140:
				case 150:
				case 160:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col + 1 == ncols) ? 0 : col + 1;
					break;
				case 180: /* far */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 200:
				case 210:
				case 220:
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 240: /* close */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 260:
				case 270:
				case 280:
					if (col == 0)
						col = ncols - 2;
					else if ((col - 1) == 0)
						col = ncols - 1;
					else
						col = col - 2;
					break;
				case 300: /* far */
					row = (row == 0) ? nrows - 1 : row - 1;
					if (col == 0)
						col = ncols - 2;
					else if ((col - 1) == 0)
						col = ncols - 1;
					else
						col = col - 2;
					break;
				case 320:
				case 330:
				case 340:
					row = (row == 0) ? nrows - 1 : row - 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
			}
			break;
		case 4:
		case 6:
			switch (dir) {
			case 0:
				col = (col + 1 == ncols) ? 0 : col + 1;
				break;
			case 45:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 60:
				if ((row & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 90:
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 120:
				if ((row & 1) != 0)
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 135:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row == 0) ? nrows - 1 : row - 1;
				break;
			case 180:
				col = (col == 0) ? ncols - 1 : col - 1;
				break;
			case 225:
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 240:
				if ((row & 1) != 0)
				col = (col == 0) ? ncols - 1 : col - 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 270:
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 300:
				if ((row & 1) == 0)
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			case 315:
				col = (col + 1 == ncols) ? 0 : col + 1;
				row = (row + 1 == nrows) ? 0 : row + 1;
				break;
			default:
				System.out.println("wrong direction " + dir);
			}
			break;
		case 5:
			int orient = ((row & 1) * 2 + col) % 4;
			switch (orient) {
			case 0:
				switch (dir) {
				case 0:
					col++;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col + 2 >= ncols) ? 0 : col + 2;
					break;
				case 102: /* 7 corner */
					col = (col + 3 >= ncols) ? 1 : col + 3;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col++;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 255: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 1:
				switch (dir) {
				case 0:
					col--;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col == 1) ? ncols - 1 : col - 2;
					break;
				case 102: /* 7 */
					col = (col == 1) ? ncols - 2 : col - 3;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col--;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 255: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 2:
				switch (dir) {
				case 0:
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					row = (row == 0) ? nrows - 1 : row - 1;
					col++;
					break;
				case 102: /* 7 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col = (col == 0) ? ncols - 2 : col - 2;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 255: /* 7 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					col = (col == 0) ? ncols - 1 : col - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			case 3:
				switch (dir) {
				case 0:
					col--;
					break;
				case 51: /* 7 */
				case 72: /* 5 */
					col = (col == 0) ? ncols - 1 : col - 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 102: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row + 1 == nrows) ? 0 : row + 1;
					break;
				case 144: /* 5 */
				case 153: /* 7 */
					col = (col + 2 >= ncols) ? 1 : col + 2;
					break;
				case 204: /* 7 */
				case 216: /* 5 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					break;
				case 255: /* 7 */
					col = (col + 1 >= ncols) ? 0 : col + 1;
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				case 288: /* 5 */
				case 306: /* 7 */
					row = (row == 0) ? nrows - 1 : row - 1;
					break;
				default:
					System.out.println("wrong direction " + dir);
				}
				break;
			default:
				System.out.println("wrong orient " + orient);
			}
			break;
		default:
			break;
		}
		return getPosition(col, row);
	}

	byte incCell(int col, int row) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);
		byte nextState = (byte) ((state < param.getNColors() - 1) ?
			state + 1 : 0);

		field.setCell(p, nextState);
		if (state < param.getNColors() - 1) {
			nCells++;
			checkAliveBounds(col, row);
		} else {
			nCells--;
		}
		CellInfo info = new CellInfo(p, nextState, (byte) 0);
		drawCell(info);
		((Ant2DFrame) frame).updateCells(nCells);
		return nextState;
	}

	int cellCount() {
		int count = 0;

		for (int j = minRow; j <= maxRow; j++)
		for (int i = minCol; i <= maxCol; i++) {
			int p = getPosition(i, j);

			if (field.getCell(p) != 0)
				count++;
		}
		return count;
	}

	boolean setCell(int col, int row, byte nextState, boolean draw) {
		int p = getPosition(col, row);
		byte state = field.getCell(p);

		if (state == 0 && nextState != 0) {
			field.setCell(p, nextState);
			nCells++;
			checkAliveBounds(col, row);
		} else if (state == nextState) {
			return false;
		} else if (nextState == 0) { // state != 0
			field.setCell(p, (byte) 0);
			nCells--;
		} else { // different sort of live cell
			field.setCell(p, nextState);
		}
		if (draw) {
			CellInfo info = new CellInfo(p, nextState, (byte) 0);

			drawCell(info);
			if ((Ant2DFrame) frame != null)
				((Ant2DFrame) frame).updateCells(nCells);
		}
		return true;
	}

	/*void printState(int state) {
		CellList curr;
		int i = 0;

		curr = first[state].next;
		System.out.println("state " + state);
		while (curr != list[state]) {
			System.out.println(i + ": position " +
				curr.info.getPosition() + ",	age " +
				curr.info.getAge() + ", state " +
				curr.info.getState() + ", toggle " +
				curr.info.getToggle());
			curr = curr.next;
			i++;
		}
	}*/

	static Color colorWheel(int state) {
		switch (state) {
		case 0: return Color.black;
		case 1: return Color.red;
		case 2: return Color.yellow;
		case 3: return Color.green;
		case 4: return Color.cyan;
		case 5: return Color.blue;
		case 6: return Color.magenta;
		case 7: return Color.pink; // orange and yellow too close
		default: return Color.white;
		}
	}

	void drawTurmite(short direction, int col, int row) {
		Graphics g = this.getGraphics();

		if (g == null)
			return;
		cellPl.drawCell(g, false, Color.white,
			(grid) ? gridColor : background, col, row);
		if (eyes && !(useTable && (neighbors == 5 ||
				(neighbors == 6 && sharpTurn))) &&
				neighbors != 7 && neighbors <= 8 &&
				neighbors >= 3) {
			cellPl.drawEyes(g, Color.black, direction, col, row);
		}
	}

	void drawCell(CellInfo info) {
		Graphics g = this.getGraphics();
		Color color;
		int col = getColumn(info);
		int row = getRow(info);
		byte state = info.getCell();

		if (state == 0) {
			color = background;
			nCells--;
		} else {
			if (MONO) {
				color = Color.white;
			} else {
				color = colorWheel(state);
			}
			nCells++;
		}
		cellPl.drawCell(g, false, color, (grid) ? gridColor : background,
			col, row);
	}

	Ant2DMatrix getField() {
		return field;
	}

class Capturer {
	private Point anchor = new Point(0, 0);
	private Point movable = new Point(0, 0);
	private Point start = new Point(0, 0);
	private Point stop = new Point(0, 0);
	private int completed = -1;
	private final Color selectedColor = Color.magenta;

	Point calcPoint(int x, int y) {
		Point t = new Point(x, y);

		if (t.x < 0)
			t.x = 0;
		if (t.y < 0)
			t.y = 0;
		if (t.x >= ncols)
			t.x = ncols - 1;
		if (t.y >= nrows)
			t.y = nrows - 1;
		return t;
	}

	void calcStartPoint() {
		int polygon = getPolygon();

		if (anchor.x > movable.x)
			start.x = movable.x;
		else
			start.x = anchor.x;
		if (anchor.y > movable.y)
			start.y = movable.y;
		else
			start.y = anchor.y;
		/* System.out.println("calcStartPoint: " +
			start.x + " " + start.y); */
		if (polygon == 3) {
			if (((start.x + start.y) & 1) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x--;
				}
			}
		} else if (polygon == 5) {
			int mod = (start.x + 2 * start.y) % 4;

			if ((mod / 2) == 1) {
				if ((start.y & 1) == 1) {
					start.y--;
				} else {
					start.x -= 2 ;
				}
			}
			start.x -= (mod & 1);
		} else if ((polygon == 6) && ((start.y & 1) == 1)) {
			start.y--;
		}
	}

	void calcEndPoint() {
		int polygon = getPolygon();

		if (anchor.x < movable.x)
			stop.x = movable.x;
		else
			stop.x = anchor.x;
		if (anchor.y < movable.y)
			stop.y = movable.y;
		else
			stop.y = anchor.y;
		if (polygon == 3) {
			if (((stop.x + stop.y) & 1) == 1) {
				if ((stop.y & 1) == 0)
					stop.y++;
				else
					stop.x++;
			}
		} else if (polygon == 5) {
			int mod = 3 - ((stop.x + 2 * stop.y) % 4);

			if ((mod / 2) == 1) {
				if ((stop.y & 1) == 0) {
					stop.x += 2;
				} else {
					stop.y++;
				}
			}
			stop.x += (mod & 1);
		} else if ((polygon == 6) && ((stop.y & 1) == 0)) {
			stop.y++;
		}
	}

	void startCapture(int column, int row) {
		movable = anchor = calcPoint(column, row);
		calcStartPoint();
		calcEndPoint();
		completed = 0;
		drawCaptureRect((grid) ? gridColor : background);
	}

	void continueCapture(int column, int row) {
		int captureColumns = getColumns();
		int captureRows = getRows();

		movable = calcPoint(column, row);
		drawCaptureRect((grid) ? gridColor : background);
		if (column >= 0 && row >= 0 &&
				column < captureColumns && row < captureRows) {
			calcStartPoint();
			calcEndPoint();
			drawCaptureRect(selectedColor);
			return;
		}
		calcStartPoint();
		calcEndPoint();
	}

	void stopCapture() {
		if (completed == 0)
			completed = 1;
	}

	void clearCapture() {
		drawCaptureRect((grid) ? gridColor : background);
		anchor.setLocation(0, 0);
		movable.setLocation(0, 0);
		completed = -1;
	}

	private void drawCaptureRect(Color color) {
		Graphics g = getGraphics();
		int x = start.x;
		int y = start.y;
		int width = stop.x - start.x + 1;
		int height = stop.y - start.y + 1;

		if (completed == -1)
			return;
		cellPl.drawCaptureRect(g, color,
			x, y, width, height);
	}

	void show() {
		if (!anchor.equals(movable))
			drawCaptureRect(selectedColor);
	}

	Rectangle getRectangle() {
		if (completed == 1) {
			int x = start.x;
			int y = start.y;
			int width = stop.x - start.x + 1;
			int height = stop.y - start.y + 1;

			/*System.out.println("getRectangle: " +
				x + " " + y + " " + width + " " + height); */
			return new Rectangle(x, y, width, height);
		}
		return null;
	}

}

}
