package com.sillycycle.bagleyd.ant2d;

/*
 * @(#)Ant2DDialog.java
 *
 * Structure from Java Sourcebook by Ed Anuff, pp 287-289
 *
 * Copyright 2010 - 2018  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.GridLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.Slider;

/**
 * The <code>Ant2DDialog</code> class holds the configuration
 * setup for the <code>Ant2DFrame</code>.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Ant2DDialog extends Dialog {
	private static final long serialVersionUID = 42L;
	static boolean active = false;
	static Ant2DFrame frame = null;

	static int columnsValue = Ant2DInterface.DEFAULT_COLUMNS;
	static int neighborIndexValue = Ant2DInterface.DEFAULT_NEIGHBOR_INDEX;
	static boolean randomOrientValue = false;
	static boolean gridValue = true;
	static boolean wrapValue = true;
	static boolean visibleValue = true;
	static boolean useTableValue = false;
	static boolean eyesValue = true;
	static boolean sharpTurnValue = false;
	static boolean truchetValue = true;

	Slider columnsChanger = new Slider("Field Size",
		columnsValue, Ant2DInterface.MIN_COLUMNS,
		Ant2DInterface.MAX_COLUMNS);
	Ant2DBoxes antBoxes = new Ant2DBoxes();
	JCheckBox randomOrientCheckBox = new JCheckBox("Random Orient");
	JCheckBox gridCheckBox = new JCheckBox("Grid");
	JCheckBox wrapCheckBox = new JCheckBox("Wrap");
	JCheckBox visibleCheckBox = new JCheckBox("Visible");
	JCheckBox useTableCheckBox = new JCheckBox("UseTable");
	JCheckBox eyesCheckBox = new JCheckBox("Eyes");
	JCheckBox sharpTurnCheckBox = new JCheckBox("SharpTurn");
	JCheckBox truchetCheckBox = new JCheckBox("Truchet");

	public Ant2DDialog(Ant2DFrame frame, Frame parent,
			String titleText,
			String buttonText) {
		super(parent, titleText, true);

		JButton button = new JButton(buttonText);
		JPanel buttonPanel = new JPanel();
		JPanel sliderPanel = new JPanel();
		JPanel antBoxesPanel = new JPanel();
		JPanel togglePanel = new JPanel();

		columnsChanger.setWidth(150);
		//ruleChanger.setWidth(256);
		buttonPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.CENTER));
		sliderPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		sliderPanel.add(columnsChanger, BorderLayout.PAGE_START);
		//sliderPanel.add(ruleChanger, BorderLayout.PAGE_START);
		buttonPanel.add(sliderPanel, BorderLayout.PAGE_START);
		antBoxesPanel.add(antBoxes, BorderLayout.PAGE_START);
		buttonPanel.add(antBoxesPanel, BorderLayout.PAGE_START);
		togglePanel.setLayout(new GridLayout(0, 2));
		randomOrientCheckBox.setSelected(randomOrientValue);
		togglePanel.add(randomOrientCheckBox);
		gridCheckBox.setSelected(gridValue);
		togglePanel.add(gridCheckBox);
		wrapCheckBox.setSelected(wrapValue);
		togglePanel.add(wrapCheckBox);
		visibleCheckBox.setSelected(visibleValue);
		togglePanel.add(visibleCheckBox);
		useTableCheckBox.setSelected(useTableValue);
		togglePanel.add(useTableCheckBox);
		eyesCheckBox.setSelected(eyesValue);
		togglePanel.add(eyesCheckBox);
		sharpTurnCheckBox.setSelected(sharpTurnValue);
		togglePanel.add(sharpTurnCheckBox);
		truchetCheckBox.setSelected(truchetValue);
		togglePanel.add(truchetCheckBox);
		buttonPanel.add(togglePanel, BorderLayout.PAGE_START);

		buttonPanel.add(button, BorderLayout.PAGE_START);
		add(BorderLayout.PAGE_START, buttonPanel);
		pack();
		Ant2DDialog.active = true;
		Ant2DDialog.frame = frame;
		columnsChanger.addChangeListener(new ColumnsChangeListener());
		antBoxes.addActionListener(new AntActionListener());
		randomOrientCheckBox.addItemListener(new RandomOrientListener());
		gridCheckBox.addItemListener(new GridListener());
		wrapCheckBox.addItemListener(new WrapListener());
		visibleCheckBox.addItemListener(new VisibleListener());
		useTableCheckBox.addItemListener(new UseTableListener());
		eyesCheckBox.addItemListener(new EyesListener());
		sharpTurnCheckBox.addItemListener(new SharpTurnListener());
		truchetCheckBox.addItemListener(new TruchetListener());

		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});

		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent event) {
				dispose();
				//else multiple Dialogs can be opened
				active = false;
			}
		});
	}

	class ColumnsChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = columnsValue;

			columnsValue = columnsChanger.getValue();
			if (oldValue != columnsValue) {
				frame.endGame();
				frame.setColumns(columnsValue);
				frame.startGame();
			}
		}
	}

	class AntActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = antBoxes.getNeighborIndex();
//System.out.println("actionPerformed: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class AntItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			int ni = neighborIndexValue;

			neighborIndexValue = antBoxes.getNeighborIndex();
//System.out.println("itemStateChanged: " + ni + " " + neighborIndexValue);
			frame.setRules(ni, neighborIndexValue);
		}
	}

	class GridListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setGrid(gridCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class RandomOrientListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setRandomOrient(randomOrientCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class WrapListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setWrap(wrapCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class VisibleListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setCellVisible(visibleCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class UseTableListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setUseTable(useTableCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class EyesListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setEyes(eyesCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	class SharpTurnListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setSharpTurn(sharpTurnCheckBox.isSelected());
			frame.requestFocus();
		}
	}
	class TruchetListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent e) {
			frame.setTruchet(truchetCheckBox.isSelected());
			frame.requestFocus();
		}
	}

	public void setColumns(int columnsValue) {
		Ant2DDialog.columnsValue = columnsValue;
		columnsChanger.setValue(columnsValue);
	}

	public static int getColumns() {
		return columnsValue;
	}

	public void setNeighborIndex(int neighborIndexValue) {
		Ant2DDialog.neighborIndexValue = neighborIndexValue;
		antBoxes.setNeighborIndex(neighborIndexValue);
	}

	public static int getNeighborIndex() {
		return neighborIndexValue;
	}

	public void setRandomOrient(boolean flag) {
		Ant2DDialog.randomOrientValue = flag;
		randomOrientCheckBox.setSelected(flag);
	}

	public static boolean getRandomOrient() {
		return randomOrientValue;
	}

	public void setGrid(boolean flag) {
		Ant2DDialog.gridValue = flag;
		gridCheckBox.setSelected(flag);
	}

	public static boolean getGrid() {
		return gridValue;
	}

	public void setWrap(boolean flag) {
		Ant2DDialog.wrapValue = flag;
		wrapCheckBox.setSelected(flag);
	}

	public static boolean getWrap() {
		return wrapValue;
	}

	public void setCellVisible(boolean flag) {
		Ant2DDialog.visibleValue = flag;
		visibleCheckBox.setSelected(flag);
	}

	public static boolean getCellVisible() {
		return visibleValue;
	}

	public void setUseTable(boolean flag) {
		Ant2DDialog.useTableValue = flag;
		useTableCheckBox.setSelected(flag);
	}

	public static boolean getUseTable() {
		return useTableValue;
	}

	public void setEyes(boolean flag) {
		Ant2DDialog.eyesValue = flag;
		eyesCheckBox.setSelected(flag);
	}

	public static boolean getEyes() {
		return eyesValue;
	}

	public void setSharpTurn(boolean flag) {
		Ant2DDialog.sharpTurnValue = flag;
		sharpTurnCheckBox.setSelected(flag);
	}

	public static boolean getSharpTurn() {
		return sharpTurnValue;
	}

	public void setTruchet(boolean flag) {
		Ant2DDialog.truchetValue = flag;
		truchetCheckBox.setSelected(flag);
	}

	public static boolean getTruchet() {
		return truchetValue;
	}

	public static boolean getActive() {
		return Ant2DDialog.active;
	}
}
