package com.sillycycle.bagleyd.ant2d;

/* Code for selection spliced in is due to Alex Shvets */
/* http://home.comcast.net/~shvets/Programs/Examples/GameLife/life2.html */
// Ant2DMatrix.java

import java.awt.Rectangle;
import java.io.Externalizable;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

public class Ant2DMatrix implements Externalizable {
	private static final long serialVersionUID = 42L;

	private int columns, rows;
	private int area;

	private byte[][] cells;

	// must exist for deserialization process
	public Ant2DMatrix() {
	}

	public Ant2DMatrix(int columns, int rows) {
		this.columns = columns;
		this.rows = rows;
		area = columns * rows;
		init();
	}

	public final void init() {
		cells = new byte[2][area]; /* state and truchet */
		clear();
	}

	public void clear() {
		for (int k = 0; k < 2; k++)
			for (int position = 0; position < area; position++) {
				cells[k][position] = (byte) 0; // Empty
			}
	}

	public int getRows() {
		return rows;
	}

	public int getColumns() {
		return columns;
	}

	public byte getCell(int position) {
		return cells[0][position];
	}

	public void setCell(int position, byte cell) {
		cells[0][position] = cell;
	}

	public void clearCell(int position) {
		cells[0][position] = (byte) 0;
		cells[1][position] = (byte) 0;
	}

	public byte getCell(int column, int row) {
		if ((row >= 0) && (row < rows) &&
				(column >= 0) && (column < columns))
			return cells[0][row * columns + column];
		return 0;
	}

	public void setCell(int column, int row, byte cell) {
		if ((row >= 0) && (row < rows) &&
				(column >= 0) && ( column < columns))
			cells[0][row * columns + column] = cell;
	}

	public boolean trucheted(int position) {
		return (cells[1][position] != 0);
	}

	public int getTruchet(int position) {
		return cells[1][position] - 1;
	}

	public void setTruchet(int position, int truchetState) {
		cells[1][position] = (byte) (truchetState + 1);
	}

	public Ant2DMatrix getSubMatrix(Rectangle r) {
		Ant2DMatrix selection = new Ant2DMatrix(r.width, r.height);

		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			selection.setCell(i, j,
				getCell(i + r.x, j + r.y));
		}
		return selection;
	}

	public void setSubMatrix(Ant2DMatrix selection, Rectangle r) {
		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			try {
				setCell(i + r.x, j + r.y,
					selection.getCell(i, j));
			} catch (ArrayIndexOutOfBoundsException e) {
				// ignore
			}
		}
	}

	public void clearSubMatrix(Rectangle r) {
		for (int k = 0; k < 2; k++)
		for (int j = r.y; j < r.y + r.height; j++)
		for (int i = r.x; i < r.x + r.width; i++) {
			cells[k][j * columns + i] = 0; //Empty
		}
	}

	private static int ceil2(int z) {
		if (z >= 0)
			return (z + 1) / 2;
		return z / 2;
	}

	/* The center point of rectangle (more or less) is used
	 * to rotate about.
	 */
	void rotate(Ant2DMatrix selection, Rectangle r,
			int polygon, boolean cw) {
		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			switch (polygon) {
			case 3:
				if (cw)
					setCell(2 * (2 * r.height / 3) + ceil2(i - j) - j - 2 + r.x,
						(i + j) / 2 + 1 + r.y,
						selection.getCell(i, j));
				else
					setCell(j + (i + j) / 2 - 3 + r.x,
						2 * (r.width / 4) - ceil2(i - j) + r.y,
						selection.getCell(i, j));
				break;
			case 4:
				if (cw)
					setCell(r.height - j - 1 + r.x,
						i + r.y,
						selection.getCell(i, j));
				else
					setCell(j + r.x,
						r.width - i - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 5:
				int mod = (i + 2 * j) % 4;
				int tcol = j * 2;
				int trow = i / 2;

				if (mod == 1 || mod == 2)
					tcol++;
				if (cw)
					setCell(4 * (r.height / 2) - tcol - 1 + r.x,
						trow + r.y,
						selection.getCell(i, j));
				else
					setCell(tcol + r.x,
						2 * (r.width / 4) - trow - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 6:
				if (cw)
					setCell(2 * (r.height / 2) - j + (i + 1 + j / 2) / 2 + r.x,
						i + j / 2 + r.y,
						selection.getCell(i, j));
				else
					setCell(j + (i + 1 - ((j + 1) / 2)) / 2 + r.x,
						2 * (r.width / 2) - i + (j + 1) / 2 - 2 + r.y,
						selection.getCell(i, j));
				break;
			default:
			}
		}
	}

	/* Wall means you reflecting in a wall mirror,
	 * else it means you are reflecting off ground.
	 */
	void reflect(Ant2DMatrix selection, Rectangle r,
			int polygon, boolean wall) {

		for (int j = 0; j < r.height; j++)
		for (int i = 0; i < r.width; i++) {
			switch (polygon) {
			case 3:
				if (wall) {
					int evenX = 1 - (r.width & 1);

					setCell(r.width - i - 1 + evenX + r.x,
						j + r.y,
						selection.getCell(i, j));
				} else {
					int oddY = (r.height & 1);

					setCell(i + r.x,
						r.height - j - 1 + oddY + r.y,
						selection.getCell(i, j));
				}
				break;
			case 4:
				if (wall)
					setCell(r.width - i - 1 + r.x,
						j + r.y,
						selection.getCell(i, j));
				else
					setCell(i + r.x,
						r.height - j - 1 + r.y,
						selection.getCell(i, j));
				break;
			case 5:
				int mod = (i + 2 * j) % 4;

				if (wall) {
					int oddX = ((r.width / 2) & 1);
					int tcol = r.width - i;

					if (mod == 1)
						tcol += 2;
					else if (mod == 2 || mod == 3)
						tcol++;
					setCell(tcol + r.x + 2 * oddX,
						j + r.y,
						selection.getCell(i, j));
				} else {
					int oddY = (r.height & 1);
					int tcol = i;
					int trow = r.height - j + oddY;

					if (mod == 0)
						tcol++;
					else if (mod == 1)
						tcol--;
					setCell(tcol + r.x,
						trow + r.y,
						selection.getCell(i, j));
				}
				break;
			case 6:
				if (wall)
					setCell(r.width - i - 1 - (((j & 1) == 1) ? 0 : 1) + r.x,
						j + r.y,
						selection.getCell(i, j));
				else
					setCell(i + r.x,
						r.height - j + r.y - 2,
						selection.getCell(i, j));
				break;
			default:
			}
		}
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();

		for (int j = 0; j < rows; j++) {
			for (int i = 0; i < columns; i++)
				sb.append(cells[0][j * columns + i]);
			sb.append("\n");
		}
		sb.append("\n");
		return sb.toString();
	}


	// implements Externalizable

	@Override
	public void writeExternal(ObjectOutput out) throws IOException {
		out.writeByte(columns);
		out.writeByte(rows);
		for (int j = 0; j < rows; j++)
		for (int i = 0; i < columns; i++) {
			if (cells[0][j * columns + i] == 1) {
				out.writeByte(i);
				out.writeByte(j);
			}
		}
	}

	@Override
	public void readExternal(ObjectInput in)
			throws IOException, ClassNotFoundException {
		columns = in.readByte();
		rows = in.readByte();
		init();
		while (true) {
			try {
				int i = in.readByte();
				int j = in.readByte();
				cells[0][j * columns + i] = 1;
			} catch (IOException e)	{
				break;
			}
		}
	}

	public static void main(String[] args) {
		System.out.println("Constructing object:");
		Ant2DMatrix m = new Ant2DMatrix(9, 8);
		m.setCell(2, 3, (byte) 1);
		m.setCell(3, 3, (byte) 1);
		m.setCell(4, 3, (byte) 1);
		m.setCell(4, 4, (byte) 1);
		m.setCell(3, 5, (byte) 1);
		Ant2DMatrix n;

		System.out.println("Saving object:");
		try (ObjectOutputStream out = new ObjectOutputStream(
				new FileOutputStream("junk.out"))) {
			out.writeObject(m);
		} catch (Exception e) {
			e.printStackTrace();
		}
		// Now get it back
		System.out.println("Recovering object:");
		try (ObjectInputStream in = new ObjectInputStream(
				new FileInputStream("junk.out"))) {
			n = (Ant2DMatrix)in.readObject();
			System.out.println(n.toString());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}

