package com.sillycycle.bagleyd.util;

/*
 * @(#)Cell2D.java
 *
 * Copyright 2009 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

/**
 * The <code>Cell2D</code> class handles the construction of the cells.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Cell2D extends Component {
	private static final long serialVersionUID = 42L;

	CellTr cellTr = null;
	CellSq cellSq = null;
	CellPn cellPn = null;
	CellHx cellHx = null;

	int polygon; /* sides of polygon = 3, 4, 5, or 6 */

	public Cell2D(int polygon, int nx, int ny,
			int screenWidth, int screenHeight, boolean scroll) {
		cellTr = null;
		cellSq = null;
		cellPn = null;
		cellHx = null;
		this.polygon = polygon;
		switch (polygon) {
		case 3:
			cellTr = new CellTr(nx, ny,
				screenWidth, screenHeight, scroll);
			break;
		case 4:
			cellSq = new CellSq(nx, ny,
				screenWidth, screenHeight, scroll);
			break;
		case 5:
			cellPn = new CellPn(nx, ny,
				screenWidth, screenHeight, scroll);
			break;
		case 6:
			cellHx = new CellHx(nx, ny,
				screenWidth, screenHeight, scroll);
			break;
		default:
		}
	}

	public Rectangle boundRect(int nx, int ny) {
		switch (polygon) {
		case 3:
			return cellTr.boundRect(nx, ny);
		case 4:
			return cellSq.boundRect(nx, ny);
		case 5:
			return cellPn.boundRect(nx, ny);
		case 6:
			return cellHx.boundRect(nx, ny);
		default:
			return null;
		}
	}

	public Point screenToCell(int x, int y) {
		switch (polygon) {
		case 3:
			return cellTr.screenToCell(x, y);
		case 4:
			return cellSq.screenToCell(x, y);
		case 5:
			return cellPn.screenToCell(x, y);
		case 6:
			return cellHx.screenToCell(x, y);
		default:
			return null;
		}
	}

	public void drawCell(Graphics g, boolean round, Color cid, Color bid,
			int i, int j) {
		if (g == null)
			return;
		switch (polygon) {
		case 3:
			cellTr.drawCell(g, round, cid, bid, i, j);
			break;
		case 4:
			cellSq.drawCell(g, round, cid, bid, i, j);
			break;
		case 5:
			cellPn.drawCell(g, round, cid, bid, i, j);
			break;
		case 6:
			cellHx.drawCell(g, round, cid, bid, i, j);
			break;
		default:
		}
	}

	public void drawCaptureRect(Graphics g, Color color,
			int i, int j, int nx, int ny) {
		if (g == null)
			return;
		switch (polygon) {
		case 3:
			cellTr.drawCaptureRect(g, color, i, j, nx, ny);
			break;
		case 4:
			cellSq.drawCaptureRect(g, color, i, j, nx, ny);
			break;
		case 5:
			cellPn.drawCaptureRect(g, color, i, j, nx, ny);
			break;
		case 6:
			cellHx.drawCaptureRect(g, color, i, j, nx, ny);
			break;
		default:
		}
	}

	public void drawEyes(Graphics g, Color color, short direction,
			int i, int j) {
		if (g == null)
			return;
		switch (polygon) {
		case 3:
			cellTr.drawEyes(g, color, direction, i, j);
			return;
		case 4:
			cellSq.drawEyes(g, color, direction, i, j);
			return;
		case 5:
			cellPn.drawEyes(g, color, direction, i, j);
			return;
		case 6:
			cellHx.drawEyes(g, color, direction, i, j);
			break;
		default:
		}
	}

	public void drawTruchet(Graphics g, Color color, int truchetState,
			boolean sharpTurn, int i, int j) {
		if (g == null)
			return;
		switch (polygon) {
		case 3:
			cellTr.drawTruchet(g, color, truchetState, i, j);
			return;
		case 4:
			cellSq.drawTruchet(g, color, truchetState, i, j);
			return;
		case 5:
			cellPn.drawTruchet(g, color, truchetState,
				sharpTurn, i, j);
			return;
		case 6:
			cellHx.drawTruchet(g, color, truchetState,
				sharpTurn, i, j);
			break;
		default:
		}
	}
}
