package com.sillycycle.bagleyd.util;

/*
 * Similar to CheckboxGroup but allow more than one on and have one label.
 * Beginnings from Core Web Programming 1st Ed. by Marty Hall and Larry Brown
 * http://www.apl.jhu.edu/~hall/CWP-Chapter13/CWP-Chapter13.4.html
 * Rest like the mask added by David Bagley
 */

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

/**
 * The <code>CheckBoxes</code> class combines an array of CheckBoxes and
 * a Label in one widget.
 *
 * @author Marty Hall and Larry Brown
 * @author David Bagley
 */

public class CheckBoxes extends JPanel {
	private static final long serialVersionUID = 42L;
	JPanel boxPanel;
	JCheckBox[] checkBox;
	JLabel label;
	String string;
	static final int MAX_SIZE = 30; // bit limit on int, consider using long
	int size = MAX_SIZE;
	int value, number;
	boolean reverse = false;

	public CheckBoxes(String labelString, int n, int bits, int max) {
		int i;

		//super("CheckBoxes");
		//setFont(new Font("Helvetica", Font.BOLD, 18));
		setLayout(new BorderLayout());
		string = "  " + labelString;
		value = bits;
		number = n;
		if (max > 0 && max <= MAX_SIZE)
			size = max;
		//label = new JLabel("0x" +
		//	Integer.toHexString(value).toUpperCase() +
		//	string, SwingConstants.CENTER);
		label = new JLabel(ruleBuilder(value) + string,
			SwingConstants.CENTER);
		add(label, BorderLayout.PAGE_START);
		boxPanel = new JPanel();
		boxPanel.setLayout(new GridLayout(3, size / 3));
		checkBox = new JCheckBox[size + 1];
		for (i = ((reverse) ? size : 0);
				((reverse) ? (i >= 0) : (i <= size));
				i += ((reverse) ? (-1) : 1)) {
			checkBox[i] = new JCheckBox(Integer.toString(i));
			if ((value & (1 << i)) != 0) {
				checkBox[i].setSelected(true);
			}
			boxPanel.add(checkBox[i]);
		}
		add(boxPanel, BorderLayout.CENTER);
		//pack();
		//show();
		for (i = 0; i <= size; i++) {
			checkBox[i].addItemListener((ItemEvent e) -> {
				updateMask();
				//label.setText("0x" +
				//	Integer.toHexString(value).toUpperCase() +
				//	string);
				label.setText(ruleBuilder(value) + string);
			});
		}
	}

	synchronized void updateMask() {
		int i, n = 0;

		for (i = 0; i <= size; i++) {
			if (i > number) {
				if (checkBox[i].isSelected())
					checkBox[i].setSelected(false);
			} else {
				if (checkBox[i].isSelected())
					n |= (1 << i);
			}
		}
		value = n;
	}

	public void addItemListener(ItemListener l) {
		for (int i = 0; i <= size; i++) {
			checkBox[i].addItemListener(l);
		}
	}

	public synchronized int getValue() {
		updateMask();
		return value;
	}

	public synchronized void setValue(int val) {
		for (int i = 0; i <= size; i++) {
			checkBox[i].setSelected((val & (1 << i)) != 0);
		}
	}

	public final String ruleBuilder(int val) {
		StringBuilder rule = new StringBuilder();

		for (int i = 0; i <= size; i++) {
			if ((val & (1 << i)) != 0) {
				//if (i >= 10)
				//	rule.append((char) ('A' + (char) (i - 10)));
				if (i >= 10)
					rule.append("_").append(i);
				else
					rule.append(i);
			}
		}
		return rule.toString();
	}

	public void updateBounds(int n) {
		number = Math.abs(n);
	}
}
